/*
 * Copyright (C) 1997-2007 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file vss.c the virtual streaming system
 *
 * This contains the audio streaming code of para_server which is independent
 * of the current audio format, audio file selector and of the activated
 * senders.
 */

#include "server.h"
#include <sys/mman.h> /* mmap */
#include <sys/time.h> /* gettimeofday */
#include "server.cmdline.h"
#include "afs_common.h"
#include "vss.h"
#include "send.h"
#include "error.h"
#include "string.h"
#include "fd.h"

extern const char *status_item_list[];

static struct timeval announce_tv;
static struct timeval data_send_barrier;
static struct timeval eof_barrier;
static struct timeval autoplay_barrier;

extern struct misc_meta_data *mmd;
extern struct audio_file_selector selectors[];
extern struct sender senders[];

static int audio_file;
static char *map;

#if 1
	void mp3_init(struct audio_format_handler *);
#endif

#ifdef HAVE_OGGVORBIS
	void ogg_init(struct audio_format_handler *);
#endif
#ifdef HAVE_FAAD
	void aac_afh_init(struct audio_format_handler *);
#endif

/**
 * the list of supported  audio formats
 */
static struct audio_format_handler afl[] = {
#if 1
	{
		.name = "mp3",
		.init = mp3_init,
	},
#endif
#ifdef HAVE_OGGVORBIS
	{
		.name = "ogg",
		.init = ogg_init,
	},
#endif
#ifdef HAVE_FAAD
	{
		.name = "aac",
		.init = aac_afh_init,
	},
#endif
	{
		.name = NULL,
	}
};

/** iterate over each supported audio format */
#define FOR_EACH_AUDIO_FORMAT(i) for (i = 0; afl[i].name; i++)



/**
 * check if vss status flag \a P (playing) is set
 *
 * \return greater than zero if playing, zero otherwise.
 *
 */
unsigned int vss_playing(void)
{
	return mmd->new_vss_status_flags & VSS_PLAYING;
}

/**
 * check if \a N (next) status flag is set
 *
 * \return greater than zero if set, zero if not.
 *
 */
unsigned int vss_next(void)
{
	return mmd->new_vss_status_flags & VSS_NEXT;
}

/**
 * check if a reposition request is pending
 *
 * \return greater than zero if true, zero otherwise.
 *
 */
unsigned int vss_repos(void)
{
	return mmd->new_vss_status_flags & VSS_REPOS;
}

/**
 * check if the vss is currently paused
 *
 * \return greater than zero if paused, zero otherwise.
 *
 */
unsigned int vss_paused(void)
{
	return !(mmd->new_vss_status_flags & VSS_NEXT)
		&& !(mmd->new_vss_status_flags & VSS_PLAYING);
}

/**
 * get the name of the given audio format
 * \param i the audio format number
 *
 * This returns a pointer to statically allocated memory so it
 * must not be freed by the caller.
 */
const char *audio_format_name(int i)
{
	return i >= 0?  afl[i].name : "(none)";
}

/**
 * initialize the virtual streaming system
 *
 * Call the init functions of all supported audio format handlers and
 * initialize all supported senders.
 */
void vss_init(void)
{
	int i;
	char *hn = para_hostname(), *home = para_homedir();
	long unsigned announce_time = conf.announce_time_arg > 0?
			conf.announce_time_arg : 300,
		autoplay_delay = conf.autoplay_delay_arg > 0?
			conf.autoplay_delay_arg : 0;

	PARA_DEBUG_LOG("supported audio formats: %s\n",
		SUPPORTED_AUDIO_FORMATS);
	for (i = 0; afl[i].name; i++) {
		PARA_NOTICE_LOG("initializing %s handler\n",
			afl[i].name);
		afl[i].init(&afl[i]);
	}
	ms2tv(announce_time, &announce_tv);
	PARA_INFO_LOG("announce timeval: %lums\n", tv2ms(&announce_tv));
	for (i = 0; senders[i].name; i++) {
		PARA_NOTICE_LOG("initializing %s sender\n", senders[i].name);
		senders[i].init(&senders[i]);
	}
	free(hn);
	free(home);
	if (conf.autoplay_given) {
		struct timeval now, tmp;
		mmd->vss_status_flags |= VSS_PLAYING;
		mmd->new_vss_status_flags |= VSS_PLAYING;
		gettimeofday(&now, NULL);
		ms2tv(autoplay_delay, &tmp);
		tv_add(&now, &tmp, &autoplay_barrier);
	}
}

static int get_file_info(int i)
{
	return  afl[i].get_file_info(map, mmd->size, &mmd->afi);
}

/**
 * guess the audio format judging from filename
 *
 * \param name the filename
 *
 * \return This function returns -1 if it has no idea what kind of audio
 * file this might be. Otherwise the (non-negative) number of the audio format
 * is returned.
 */
int guess_audio_format(const char *name)
{
	int i,j, len = strlen(name);

	FOR_EACH_AUDIO_FORMAT(i) {
		for (j = 0; afl[i].suffixes[j]; j++) {
			const char *p = afl[i].suffixes[j];
			int plen = strlen(p);
			if (len < plen + 1)
				continue;
			if (name[len - plen - 1] != '.')
				continue;
			if (strcasecmp(name + len - plen, p))
				continue;
//			PARA_DEBUG_LOG("might be %s\n", audio_format_name(i));
			return i;
		}
	}
	return -1;
}

static int get_audio_format(int omit)
{
	int i;

	FOR_EACH_AUDIO_FORMAT(i) {
		if (i == omit)
			continue;
		if (get_file_info(i) > 0)
			return i;
	}
	return -E_AUDIO_FORMAT;
}

/*
 * upddate shared mem
 */
static int update_mmd(void)
{
	int i;

	i = guess_audio_format(mmd->filename);
	if (i < 0 || get_file_info(i) < 0)
		i = get_audio_format(i);
	if (i < 0)
		return i;
	mmd->audio_format = i;
	mmd->chunks_sent = 0;
	mmd->current_chunk = 0;
	mmd->offset = 0;
	mmd->events++;
	return 1;
}

static void vss_get_audio_file(void)
{
	char **sl = selectors[mmd->selector_num].get_audio_file_list(10);
	int i;
	struct stat file_status;

	if (!sl)
		goto err_out;
	for (i = 0; sl[i]; i++) {
		struct timeval now;
		PARA_INFO_LOG("trying %s\n", sl[i]);
		if (strlen(sl[i]) >= _POSIX_PATH_MAX)
			continue;
		audio_file = open(sl[i], O_RDONLY);
		if (audio_file < 0)
			continue;
		if (fstat(audio_file, &file_status) == -1 ||
				!file_status.st_size) {
			close(audio_file);
			continue;
		}
		mmd->size = file_status.st_size;
		mmd->mtime = file_status.st_mtime;
		map = para_mmap(mmd->size, PROT_READ, MAP_PRIVATE,
			audio_file, 0);
		strcpy(mmd->filename, sl[i]);
		mmd->afi.header_len = 0; /* default: no header */
		if (update_mmd() < 0) { /* invalid file */
			close(audio_file);
			munmap(map, mmd->size);
			map = NULL;
			continue;
		}
		mmd->num_played++;
		if (selectors[mmd->selector_num].update_audio_file)
			selectors[mmd->selector_num].update_audio_file(sl[i]);
		PARA_NOTICE_LOG("next audio file: %s\n", mmd->filename);
		mmd->new_vss_status_flags &= (~VSS_NEXT);
		gettimeofday(&now, NULL);
		tv_add(&now, &announce_tv, &data_send_barrier);
		goto free;
	}
err_out:
	PARA_ERROR_LOG("%s", "no valid files found\n");
	mmd->new_vss_status_flags = VSS_NEXT;
free:
	if (sl) {
		for (i = 0; sl[i]; i++)
			free(sl[i]);
		free(sl);
	}
}

static int chk_barrier(const char *bname, const struct timeval *now,
		const struct timeval *barrier, struct timeval *diff,
		int print_log)
{
	long ms;

	if (tv_diff(now, barrier, diff) > 0)
		return 1;
	ms = tv2ms(diff);
	if (print_log && ms)
		PARA_DEBUG_LOG("%s barrier: %lims left\n", bname, ms);
	return -1;
}

static void vss_next_chunk_time(struct timeval *due)
{
	struct timeval tmp;

	tv_scale(mmd->chunks_sent, &mmd->afi.chunk_tv, &tmp);
	tv_add(&tmp, &mmd->stream_start, due);
}

/*
 * != NULL: timeout for next chunk
 * NULL: nothing to do
 */
static struct timeval *vss_compute_timeout(void)
{
	static struct timeval the_timeout;
	struct timeval now, next_chunk;

	if (vss_next() && mmd->audio_format >= 0) {
		/* only sleep a bit, nec*/
		the_timeout.tv_sec = 0;
		the_timeout.tv_usec = 100;
		return &the_timeout;
	}
	gettimeofday(&now, NULL);
	if (chk_barrier("autoplay_delay", &now, &autoplay_barrier,
			&the_timeout, 1) < 0)
		return &the_timeout;
	if (chk_barrier("eof", &now, &eof_barrier, &the_timeout, 1) < 0)
		return &the_timeout;
	if (chk_barrier("data send", &now, &data_send_barrier,
			&the_timeout, 1) < 0)
		return &the_timeout;
	if (mmd->audio_format < 0 || !vss_playing() || !map)
		return NULL;
	vss_next_chunk_time(&next_chunk);
	if (chk_barrier(afl[mmd->audio_format].name, &now, &next_chunk,
			&the_timeout, 0) < 0)
		return &the_timeout;
	/* chunk is due or bof */
	the_timeout.tv_sec = 0;
	the_timeout.tv_usec = 0;
	return &the_timeout;
}

static void vss_eof(void)
{
	struct timeval now;
	int i;
	char *tmp;

	if (!map) {
		for (i = 0; senders[i].name; i++)
			senders[i].shutdown_clients();
		return;
	}
	gettimeofday(&now, NULL);
	tv_add(&mmd->afi.eof_tv, &now, &eof_barrier);
	munmap(map, mmd->size);
	map = NULL;
	close(audio_file);
	mmd->audio_format = -1;
	mmd->chunks_sent = 0;
	mmd->offset = 0;
	mmd->afi.seconds_total = 0;
	free(mmd->afi.chunk_table);
	mmd->afi.chunk_table = NULL;
	tmp  = make_message("%s:\n%s:\n%s:\n", status_item_list[SI_AUDIO_INFO1],
		status_item_list[SI_AUDIO_INFO2], status_item_list[SI_AUDIO_INFO3]);
	strcpy(mmd->afi.info_string, tmp);
	free(tmp);
	tmp  = make_message("%s:\n%s:\n%s:\n", status_item_list[SI_DBINFO1],
		status_item_list[SI_DBINFO2], status_item_list[SI_DBINFO3]);
	strcpy(mmd->selector_info, tmp);
	free(tmp);
	mmd->filename[0] = '\0';
	mmd->size = 0;
	mmd->events++;
}

/**
 * get the header and of the current audio file
 *
 * \param header_len the length of the header is stored here
 *
 * \return a pointer to a buffer containing the header, or NULL, if no audio
 * file is selected or if the current audio format does not need special header
 * treamtment.
 *
 */
char *vss_get_header(size_t *header_len)
{
	if (mmd->audio_format < 0 || !map || !mmd->afi.header_len)
		return NULL;
	*header_len = mmd->afi.header_len;
	return map + mmd->afi.header_offset;
}

/**
 * get the list of all supported audio formats
 *
 * \return a space separated list of all supported audio formats
 * It is not allocated at runtime, i.e. there is no need to free
 * the returned string in the caller.
 */
const char *supported_audio_formats(void)
{
	return SUPPORTED_AUDIO_FORMATS;
}

/**
 * get the chunk time of the current audio file
 *
 * \return a pointer to a struct containing the chunk time, or NULL,
 * if currently no audio file is selected.
 */
struct timeval *vss_chunk_time(void)
{
	if (mmd->audio_format < 0)
		return NULL;
	return &mmd->afi.chunk_tv;
}

/**
 * compute the timeout for para_server's main select-loop
 *
 * This function gets called from para_server to determine the timeout value
 * for its main select loop.
 *
 * Before the timeout is computed, the current vss status flags are evaluated
 * and acted upon by calling appropriate functions from the lower layers.
 * Possible actions include
 *
 * 	- request a new file list from the current audio file selector
 *	- shutdown of all senders (stop/pause command)
 *	- reposition the stream (ff/jmp command)
 *
 * \return A pointer to a struct timeval containing the timeout for the next
 * chunk of data to be sent, or NULL if we're not sending right now.
 */
struct timeval *vss_preselect(void)
{
	struct audio_format_handler *af = NULL;
	int i, format;
	struct timeval *ret;
again:
	format = mmd->audio_format;
	if (format >= 0)
		af = afl + format;
	else
		for (i = 0; senders[i].name; i++)
			senders[i].shutdown_clients();
	if (vss_next() && af) {
		vss_eof();
		return vss_compute_timeout();
	}
	if (vss_paused() || vss_repos()) {
		for (i = 0; senders[i].name; i++)
			senders[i].shutdown_clients();
		if (af) {
			struct timeval now;
			gettimeofday(&now, NULL);
			if (!vss_paused() || mmd->chunks_sent)
				tv_add(&mmd->afi.eof_tv, &now, &eof_barrier);
			if (vss_repos())
				tv_add(&now, &announce_tv, &data_send_barrier);
			if (mmd->new_vss_status_flags & VSS_NOMORE)
				mmd->new_vss_status_flags = VSS_NEXT;
		}
		mmd->chunks_sent = 0;
	}
	if (vss_repos()) {
		mmd->new_vss_status_flags &= ~(VSS_REPOS);
		mmd->current_chunk = mmd->repos_request;
	}
	ret = vss_compute_timeout();
	if (!ret && !map && vss_playing() &&
			!(mmd->new_vss_status_flags & VSS_NOMORE)) {
		PARA_DEBUG_LOG("%s", "ready and playing, but no audio file\n");
		vss_get_audio_file();
		goto again;
	}
	return ret;
}

static void get_chunk(long unsigned chunk_num, char **buf, size_t *len)
{
	size_t pos = mmd->afi.chunk_table[chunk_num];
	*buf = map + pos;
	*len = mmd->afi.chunk_table[chunk_num + 1] - pos;
}

/**
 * Get the data of the given chunk.
 *
 * \param chunk_num The number of the desired chunk.
 * \param buf Chunk data.
 * \param len Chunk length in bytes.
 *
 * \return Positive on success, negative on errors.
 */
int vss_get_chunk(long unsigned chunk_num, char **buf, size_t *len)
{
	if (mmd->audio_format < 0 || !map || !vss_playing())
		return -E_CHUNK;
	if (chunk_num >= mmd->afi.chunks_total)
		return -E_CHUNK;
	get_chunk(chunk_num, buf, len);
	return 1;
}

/**
 * main sending function
 *
 * This function gets called from para_server as soon as the next chunk of
 * data should be pushed out. It first calls the read_chunk() function of
 * the current audio format handler to obtain a pointer to the data to be
 * sent out as well as its length. This information is then passed to each
 * supported sender's send() function which does the actual sending.
 */
void vss_send_chunk(void)
{
	int i;
	struct audio_format_handler *af;
	struct timeval now, due;

	if (mmd->audio_format < 0 || !map || !vss_playing())
		return;
	af = &afl[mmd->audio_format];
	gettimeofday(&now, NULL);
	vss_next_chunk_time(&due);
	if (tv_diff(&due, &now, NULL) > 0)
		return;
	if (chk_barrier("eof", &now, &eof_barrier, &due, 1) < 0)
		return;
	if (chk_barrier("data send", &now, &data_send_barrier,
			&due, 1) < 0)
		return;
	mmd->new_vss_status_flags &= ~VSS_REPOS;
	if (mmd->current_chunk >= mmd->afi.chunks_total) { /* eof */
		mmd->new_vss_status_flags |= VSS_NEXT;
		return vss_eof();
	}
	/*
	 * We call the send function also in case of empty chunks as they
  	 * might have still some data queued which can be sent in this case.
  	 */
	if (!mmd->chunks_sent) {
		struct timeval tmp;
		gettimeofday(&mmd->stream_start, NULL);
		tv_scale(mmd->current_chunk, &mmd->afi.chunk_tv, &tmp);
		mmd->offset = tv2ms(&tmp);
		mmd->events++;
	}
	for (i = 0; senders[i].name; i++) {
		char *buf;
		size_t len;
		get_chunk(mmd->current_chunk, &buf, &len);
		senders[i].send(mmd->current_chunk, mmd->chunks_sent, buf, len);
	}
	mmd->new_vss_status_flags |= VSS_PLAYING;
	mmd->chunks_sent++;
	mmd->current_chunk++;
}
